{----------------------------------------------------------------------------
|
| Library: Spider Containers for Object Pascal
|
| Module: ObjCmp.Pas
|
| Description: Various functions and types used in object comparison.
|
| Copyright (c) 1996,97 Michel Brazeau, Interval Software
|
| History: Version 1.0  March 1996. 
|          May 14, 1997. MB add NegateComparison
|          May 17, 1997. MB add CompareStringNoCase
|          May 18, 1997. MB add CompareDateTime
|
|---------------------------------------------------------------------------}
unit ObjCmp;

{$I Contain.Inc}

interface

type

{ comparison result value }
TComparison = (cmpEqual, cmpLess, cmpGreater);

{ Compares Lhs (left hand side) with Rhs. return cmpLess when Lhs < Rhs
  NOTE: All compare functions must be compiled in the F+ state or with the
        far directive }
TCompareFunc = function(const Lhs, Rhs : TObject) : TComparison;


{ dummy compare functions }
function CompareEqual(const Lhs, Rhs : TObject) : TComparison; far; { returns cmpEqual }
function CompareLess(const Lhs, Rhs : TObject) : TComparison; far; { returns cmpLess }
function CompareGreater(const Lhs, Rhs : TObject) : TComparison; far; { returns cmpGreater }

{ type specific compare functions that may be used within generic compare
  functions }
function CompareString(const Lhs, Rhs : String) : TComparison; far;
function CompareStringNoCase(const Lhs, Rhs : String) : TComparison; far;
function CompareReal(const Lhs, Rhs : Extended) : TComparison; far;
function CompareInteger(const Lhs, Rhs : LongInt) : TComparison; far;
function CompareDateTime( const Lhs, Rhs : TDateTime ) : TComparison; far;

{ returns the inverse of a TComparison }
function NegateComparison( Comparison : TComparison ) : TComparison;

implementation

uses
    SysUtils, { AnsiUpperCase }
    ObjPerst; { RegisterCompareFunc }

{--------------------------------------------------------------------------}

function CompareEqual(const Lhs, Rhs : TObject) : TComparison;
begin
    Result := cmpEqual;
end;

{--------------------------------------------------------------------------}

function CompareLess(const Lhs, Rhs : TObject) : TComparison;
begin
    Result := cmpLess;
end;

{--------------------------------------------------------------------------}

function CompareGreater(const Lhs, Rhs : TObject) : TComparison;
begin
    Result := cmpGreater;
end;

{--------------------------------------------------------------------------}

function CompareString(const Lhs, Rhs : String) : TComparison;
var
    CompareResult : Integer;
begin
    CompareResult := AnsiCompareStr(Lhs, Rhs);

    if CompareResult < 0 then
        Result := cmpLess
    else if CompareResult > 0 then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

{--------------------------------------------------------------------------}

function CompareStringNoCase(const Lhs, Rhs : String) : TComparison;
var
    CompareResult : Integer;
begin
    CompareResult := AnsiCompareText(Lhs, Rhs);

    if CompareResult < 0 then
        Result := cmpLess
    else if CompareResult > 0 then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

{--------------------------------------------------------------------------}

function CompareReal(const Lhs, Rhs : Extended) : TComparison;
begin
    if Lhs < Rhs then
        Result := cmpLess
    else if Lhs > Rhs then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

{--------------------------------------------------------------------------}

function CompareInteger(const Lhs, Rhs : LongInt) : TComparison;
begin
    if Lhs < Rhs then
        Result := cmpLess
    else if Lhs > Rhs then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

{-------------------------------------------------------------------------}

function CompareDateTime( const Lhs, Rhs : TDateTime ) : TComparison;
begin
    Result := CompareReal(Lhs, Rhs);
end;

{-------------------------------------------------------------------------}

function NegateComparison( Comparison : TComparison ) : TComparison;
begin
    case Comparison of

        cmpLess    : Result := cmpGreater;

        cmpGreater : Result := cmpLess;

        else
            Result := cmpEqual;

    end;
end;

{-------------------------------------------------------------------------}

initialization

    RegisterCompareFunction( 'CompareEqual', CompareEqual );
    RegisterCompareFunction( 'CompareLess', CompareLess );
    RegisterCompareFunction( 'CompareGreater', CompareGreater );

end.

