{----------------------------------------------------------------------------
|
| Library: Spider Containers for Object Pascal
|
| Module: ObjPerst.Pas
|
| Description: Classes and routines of object persistence.
|
| History: Version 1.0  Apr 27 1997.
|          May 14, 1997. MB add CloneObj
|          Oct 27, 1997. MB in TObjectStream.Create, Rewind is not invoked
|                        Add FreeStreamOnDestroy property.
|          Jun 07, 1998. MB add ObjectsEqual
|          Aug 29, 1998. MB Adapt TObjectStream.ReadObject/WriteObject to
|                        handle nil objects.
|          Nov 22, 1998. MB make GetStreamableInterface thread safe.
|
|---------------------------------------------------------------------------}
unit ObjPerst;

{$I Contain.Inc}

interface

uses
    {$ifdef Win32}
    Windows,  { for TRtlCriticalSection, EnterCriticalSection... }
    {$endif}
    ObjCmp,   { for TCompareFunc }
    ObjType,  { for THashFunc }
    Classes;  { for TReader, TWriter }

type

{ type which defines different versions of a class for streaming purposes.
  Do NOT change type. Must be 2 bytes in size. }
TClassVersion = Word;

TObjectReader = class;
TObjectWriter = class;

{ stream of objects. }
TObjectStream = class(TObject)
protected
    FStream      : TStream;

    { reader and writer objects created in ReadObject and WriteObject
      methods. }
    FWriter      : TObjectWriter;
    FReader      : TObjectReader;

    { nested level when streaming child objects }
    FNestedLevel : Integer;

    FFreeStreamOnDestroy : Boolean;

    function GetEndOfStream : Boolean;

public
    { Create an object stream from an allready constructed TStream instance.
      The TStream instance will be freed when the object stream is destroyed if
      FreeStreamOnDestroy is True (default). }
    constructor Create( const Stream : TStream );

    { Create an object stream from a file. Mode takes on the "file-mode
      constants" from Delphi. ie, fmOpenRead, fmOpenWrite, fmCreate... see the
      Delphi on-line help on "file-mode constants"}
    constructor CreateFromFile( const FileName : String;
                                const Mode     : Word);

    destructor Destroy; override;

    { Read an object at the current position in the stream. Reading beyond
      the end of the stream will raise Delphi's EReadError exception. }
    function ReadObject : TObject;

    { Write an object at the current position in the stream. Generally
      objects should always be written sequentially. }
    procedure WriteObject( const Obj : TObject );

    { rewind the stream to the beginning }
    procedure Rewind;

    { read-only property which is True when the current position in stream is
      at the end of the stream. May be used in constructs like,

      while not ObjectStream.EndOfStream do
          Obj := ReadObject;
    }
    property EndOfStream : Boolean read GetEndOfStream;

    { if True, FStream will be freed when the TObjectStream is destroyed.
      Default is True. }
    property FreeStreamOnDestroy : Boolean read FFreeStreamOnDestroy
                                           write FFreeStreamOnDestroy;


end;

{--------------------------------------------------------------------------}

{ class to provide the services to read an object instance from a stream.
  Augments Delphi's TReader by allowing the read child objects. }
TObjectReader = class(TReader)
private
    FObjectStream : TObjectStream;

public
    { ObjectStream will NOT be destroyed when the instance is destroyed }
    constructor Create( const ObjectStream : TObjectStream );

    { read a child object, belonging to the object being read. }
    function ReadChildObject : TObject;

end;

{--------------------------------------------------------------------------}

{ class to provide the services to write an object instance to a stream.
  Augments Delphi's TWriter by allowing the write child objects. }
TObjectWriter = class(TWriter)
private
    FObjectStream : TObjectStream;

public
    { ObjectStream will NOT be destroyed when the instance is destroyed }
    constructor Create( const ObjectStream : TObjectStream );

    { write a child object, belonging to the object being written }
    procedure WriteChildObject( const Obj : TObject );

end;

{--------------------------------------------------------------------------}

{ procedure types to read and write objects to streams. They are file level
  routines, as opposed to object methods, to allow streaming existing objects
  without subclassing them }
TObjectReadProc = procedure( const Obj    : TObject;
                             const Reader : TObjectReader );

TObjectWriteProc = procedure( const Obj    : TObject;
                              const Writer : TObjectWriter );

{--------------------------------------------------------------------------}

{ procedure which registers a streamable object. More than one read or write
  procedure may be specified to support multiple versions. The first read
  procedure corresponds to version 1, the second to version 2, ... By default,
  objects are written with the most recent write procedure, unless
  SetDefaultObjectWriteProc }
procedure RegisterStreamable( const InClass      : TClass;
                              const InReadProcs  : array of TObjectReadProc;
                              const InWriteProcs : array of TObjectWriteProc );

{ sets the default write procedure version for a class. Use this routine if
  objects of a certain type are to be written with an older version of a
  write procedures. May raise EClassVersionNotRegistered. }
procedure SetDefaultWriteProcVersion( const InClass        : TClass;
                                      const InClassVersion : TClassVersion );

{ procedure which registers a compare function. Compare functions need to be
  registered to stream entire containers. }
procedure RegisterCompareFunction( const FunctionName : String;
                                   const CompareFunc  : TCompareFunc );

{ procedure which registers a hash function. Hash functions need to be
  registered to stream containers which use hashing. }
procedure RegisterHashFunction( const FunctionName : String;
                                const HashFunc     : THashFunc );

{--------------------------------------------------------------------------}


{ returns the registered name of a compare function. May raise
  ECompareFuncNotRegistered. }
function GetCompareFuncName( const CompareFunc : TCompareFunc ) : String;

{ returns the registered name of a hash function. May raise
  EHashFuncNotRegistered. }
function GetHashFuncName( const HashFunc : THashFunc ) : String;

{ returns a compare function given its registered name. May raise
  ECompareFuncNotRegistered. }
function GetCompareFunc( const FunctionName : String ) : TCompareFunc;

{ returns a hash function given its registered name. May raise
  EHashFuncNotRegistered. }
function GetHashFunc( const FunctionName : String ) : THashFunc;

{ returns an TClass given a registered class name. May raise
  EClassNotRegistered. }
function GetRegisteredClass( const ClassName : String ) : TClass;

{--------------------------------------------------------------------------}

{ returns a newly constructed object, cloned from Obj }
function CloneObj( const Obj : TObject ) : TObject;

{--------------------------------------------------------------------------}

{ returns True if ObjA is equal to ObjB. This is a "deep" comparison, not
  just a comparison of the object references. }
function ObjectsEqual( const ObjA : TObject;
                       const ObjB : TObject ) : Boolean;

{--------------------------------------------------------------------------}

{ returns True if the contents of StreamA is identical to the constents of
  StreamB }
function StreamsEqual( const StreamA : TStream;
                       const StreamB : TStream ) : Boolean;

{--------------------------------------------------------------------------}


{ read an object from a file. If UseMemoryStream the file will be loaded in a
  memory stream first. This can increase performance dramatically. If the
  file is very large, this will consume a large amount memory resulting in
  the OS swapping to disk. }
function ReadObjectFromFile( const FileName        : String;
                             const UseMemoryStream : Boolean ) : TObject;

{ writes an object to a file. WARNING: when saving an object to a file, if a
  file with the same name exists, it is overwritten.  If UseMemoryStream
  the file will be written to a memory stream first. This can increase
  performance dramatically. If the file is very large, this will consume a
  large amount memory resulting in the OS swapping to disk. }
procedure WriteObjectToFile( const Obj             : TObject;
                             const FileName        : String;
                             const UseMemoryStream : Boolean );

{--------------------------------------------------------------------------}

{ Read and Write procedure for TGraphic descendants (TBitmap, TMetaFile,
  TIcon). TBitmap, TMetafile, and TIcon classes are automatically registered
  in the unit initialization. }
procedure WriteGraphic( const Obj     : TObject;
                        const Writer  : TObjectWriter ); far;

procedure ReadGraphic( const Obj     : TObject;
                       const Reader  : TObjectReader ); far;

{ Read and Write procedure for TMemoryStream class. The TMemoryStream class
  is automatically registered in the unit initialization. }
procedure WriteMemoryStream( const Obj     : TObject;
                             const Writer  : TObjectWriter ); far;

procedure ReadMemoryStream( const Obj     : TObject;
                            const Reader  : TObjectReader ); far;

{--------------------------------------------------------------------------}

implementation


uses
    SysUtils, { for fmOpenRead }
    Graphics, { for TGraphic }
    ObjBase,  { for TNodeContainer }
    ObjBuckt, { for TBucket }
    ObjList;  { for TUnOrderedList }

const

CReaderWriterBufSize = 4096;

CNilObjectStrId      = '*NilObject*';

var
    { registry of TStreamableInterface instances. Must be accessed with
      GetObjectRegistry }
    ObjectRegistry      : TNodeContainer;

    { registry of compare functions. Must be accessed with
      GetCompareFuncRegistry  }
    CompareFuncRegistry : TNodeContainer;

    { registry of hash functions. Must be accessed with
      GetHashFuncRegistry   }
    HashFuncRegistry    : TNodeContainer;


{--------------------------------------------------------------------------}

type

{ interface class which associates read and write procedures with an object
  type }
TStreamableInterface = class(TObject)
private
    FClassName               : String;
    FClass                   : TClass;
    FReadProcs               : TUnOrderedList;
    FWriteProcs              : TUnOrderedList;

    FDefaultWriteProcVersion : TClassVersion;

    procedure SetDefaultWriteProcVersion( const ClassVersion : TClassVersion );

    function GetReadProc( const ClassVersion : TClassVersion ) : TObjectReadProc;
    function GetWriteProc( const ClassVersion : TClassVersion ) : TObjectWriteProc;

public
    constructor Create( const InClass       : TClass;
                        const InReadProcs   : array of TObjectReadProc;
                        const InWriteProcs  : array of TObjectWriteProc );

    destructor Destroy; override;

    property ReadProcs[const ClassVersion : TClassVersion] : TObjectReadProc
                read GetReadProc;
    property WriteProcs[const ClassVersion : TClassVersion] : TObjectWriteProc
                read GetWriteProc;

    property RegisteredClass : TClass read FClass;

    property RegisteredClassName : String read FClassName
                                          write FClassName;

    property DefaultWriteProcVersion : TClassVersion
                read FDefaultWriteProcVersion
                write SetDefaultWriteProcVersion;

end;

{--------------------------------------------------------------------------}

function CompareStreamableInterface(const Lhs, Rhs : TObject) : TComparison; far;
begin
    Result := CompareString( (Lhs as TStreamableInterface).RegisteredClassName,
                             (Rhs as TStreamableInterface).RegisteredClassName );
end;

{--------------------------------------------------------------------------}

type

{ class which associates a compare function name with a compare function }
TCompareFuncEntry = class(TObject)
private
    FFunctionName : String;
    FCompareFunc  : TCompareFunc;

public
    constructor Create( const InFunctionName : String;
                        const InCompareFunc  : TCompareFunc );

    property FunctionName : String read FFunctionName;

    property CompareFunc  : TCompareFunc read FCompareFunc;
end;

{--------------------------------------------------------------------------}

{ compare function for TCompareFuncEntry }
function CompareCompareFuncEntry(const Lhs, Rhs : TObject) : TComparison; far;
begin
    Result := CompareString( (Lhs as TCompareFuncEntry).FunctionName,
                             (Rhs as TCompareFuncEntry).FunctionName );
end;

{--------------------------------------------------------------------------}

type

{ class which associates a hash function name with a hash function }
THashFuncEntry = class(TObject)
private
    FFunctionName : String;
    FHashFunc     : THashFunc;

public
    constructor Create( const InFunctionName : String;
                        const InHashFunc     : THashFunc );

    property FunctionName : String read FFunctionName;

    property HashFunc     : THashFunc read FHashFunc;
end;

{--------------------------------------------------------------------------}

{ compare function for THashFuncEntry }
function CompareHashFuncEntry(const Lhs, Rhs : TObject) : TComparison; far;
begin
    Result := CompareString( (Lhs as THashFuncEntry).FunctionName,
                             (Rhs as THashFuncEntry).FunctionName );
end;

{--------------------------------------------------------------------------}

{ return the object registry. If it is not yet created, it is created }
function GetObjectRegistry : TNodeContainer;
begin
    if ObjectRegistry = nil then
        ObjectRegistry := TUnOrderedList.Create( TStreamableInterface,
                                                 CompareStreamableInterface );
    Result := ObjectRegistry;
end;

{--------------------------------------------------------------------------}

{ return the compare function registry. If it is not yet created, it is
  created }
function GetCompareFuncRegistry : TNodeContainer;
begin
    if CompareFuncRegistry = nil then
        CompareFuncRegistry := TUnOrderedList.Create( TCompareFuncEntry,
                                                      CompareCompareFuncEntry );

    Result := CompareFuncRegistry;
end;

{--------------------------------------------------------------------------}

{ return the hash function registry. If it is not yet created, it is
  created }
function GetHashFuncRegistry : TNodeContainer;
begin
    if HashFuncRegistry = nil then
        HashFuncRegistry := TUnOrderedList.Create( THashFuncEntry,
                                                   CompareHashFuncEntry );

    Result := HashFuncRegistry;
end;

{--------------------------------------------------------------------------}

constructor TStreamableInterface.Create(
                        const InClass       : TClass;
                        const InReadProcs   : array of TObjectReadProc;
                        const InWriteProcs  : array of TObjectWriteProc );
var
    Index : Integer;

begin
    inherited Create;

    FClass      := InClass;
    FClassName  := InClass.ClassName;
    FReadProcs  := TUnOrderedList.Create(TBucket, CompareLess);
    FWriteProcs := TUnOrderedList.Create(TBucket, CompareLess);

    for Index := Low(InReadProcs) to High(InReadProcs) do
    begin
        FReadProcs.Insert( TBucket.CreateSetData( SizeOf(TObjectReadProc),
                                                  InReadProcs[Index] ) );
    end;

    for Index := Low(InWriteProcs) to High(InWriteProcs) do
    begin
        FWriteProcs.Insert( TBucket.CreateSetData( SizeOf(TObjectWriteProc),
                                                   InWriteProcs[Index] ) );
    end;

    { the default write procedure is the last one registered }
    FDefaultWriteProcVersion := FWriteProcs.Size;

end;

{--------------------------------------------------------------------------}

destructor TStreamableInterface.Destroy;
begin
    FWriteProcs.Free;
    FReadProcs.Free;

    inherited Destroy;
end;

{--------------------------------------------------------------------------}

procedure TStreamableInterface.SetDefaultWriteProcVersion(
                    const ClassVersion : TClassVersion );
begin
    if (ClassVersion >= 1) and
       (ClassVersion <= FWriteProcs.Size) then
        FDefaultWriteProcVersion := ClassVersion
    else
        RaiseClassVersionNotRegistered;
end;

{--------------------------------------------------------------------------}

function TStreamableInterface.GetReadProc(
                        const ClassVersion : TClassVersion ) : TObjectReadProc;
begin
    if (ClassVersion >= 1) and
       (ClassVersion <= FReadProcs.Size) then
    begin
        FReadProcs.GotoNth(ClassVersion);

        (FReadProcs.CurrentObj as TBucket).GetData(Result);
    end
    else
        RaiseClassVersionNotRegistered;
end;

{--------------------------------------------------------------------------}

function TStreamableInterface.GetWriteProc(
                        const ClassVersion : TClassVersion ) : TObjectWriteProc;
begin
    if (ClassVersion >= 1) and
       (ClassVersion <= FWriteProcs.Size) then
    begin
        FWriteProcs.GotoNth(ClassVersion);

        (FWriteProcs.CurrentObj as TBucket).GetData(Result);
    end
    else
        RaiseClassVersionNotRegistered;
end;

{--------------------------------------------------------------------------}

{$ifdef Win32}
var
    GetStreamableInterfaceCS : TRtlCriticalSection;
{$endif}

{ returns a TStreamableInterface given a class name. May raise
  EClassNotRegistered if ClassName is not registered. }
function GetStreamableInterface( const ClassName : String ) : TStreamableInterface;
var
    Key : TStreamableInterface;
begin
    {$ifdef Win32}
    EnterCriticalSection(GetStreamableInterfaceCS);
    try
    {$endif}

    { assign value to result to prevent compiler warning }
    Result := nil;

    Key := TStreamableInterface.Create( TObject, [nil], [nil] );
    try
        Key.RegisteredClassName := ClassName;

        if GetObjectRegistry.Search( Key, 1 ) then
            Result := (GetObjectRegistry.CurrentObj as TStreamableInterface)
        else
            RaiseClassNotRegistered(ClassName);
    finally
        Key.Free;
    end;

    {$ifdef Win32}
    finally
        LeaveCriticalSection(GetStreamableInterfaceCS);
    end;
    {$endif}
end;

{--------------------------------------------------------------------------}

function GetRegisteredClass( const ClassName : String ) : TClass;
begin
    Result := GetStreamableInterface(ClassName).RegisteredClass;
end;

{--------------------------------------------------------------------------}

procedure SetDefaultWriteProcVersion( const InClass        : TClass;
                                      const InClassVersion : TClassVersion );
var
    StreamableInterface : TStreamableInterface;
begin
    StreamableInterface := GetStreamableInterface( InClass.ClassName );

    StreamableInterface.SetDefaultWriteProcVersion( InClassVersion );
end;

{--------------------------------------------------------------------------}

procedure RegisterStreamable( const InClass      : TClass;
                              const InReadProcs  : array of TObjectReadProc;
                              const InWriteProcs : array of TObjectWriteProc );
var
    StreamableInterface : TStreamableInterface;

begin
    StreamableInterface := TStreamableInterface.Create( InClass,
                                                        InReadProcs,
                                                        InWriteProcs );

    if GetObjectRegistry.Search( StreamableInterface, 1 ) then
    begin
        StreamableInterface.Free;
        RaiseDuplicateRegisteredClass(InClass.ClassName);
    end
    else
        GetObjectRegistry.Insert(StreamableInterface);
end;

{--------------------------------------------------------------------------}

constructor TCompareFuncEntry.Create(
                        const InFunctionName : String;
                        const InCompareFunc  : TCompareFunc );
begin
    FFunctionName := InFunctionName;
    FCompareFunc  := InCompareFunc;

    inherited Create;
end;

{--------------------------------------------------------------------------}

procedure RegisterCompareFunction( const FunctionName : String;
                                   const CompareFunc  : TCompareFunc );
var
    CompareFuncEntry : TCompareFuncEntry;

begin
    CompareFuncEntry := TCompareFuncEntry.Create(FunctionName, CompareFunc);

    if GetCompareFuncRegistry.Search( CompareFuncEntry, 1 ) then
    begin
        CompareFuncEntry.Free;
        RaiseDuplicateCompareFunc(FunctionName);
    end
    else
        GetCompareFuncRegistry.Insert( CompareFuncEntry );
end;

{--------------------------------------------------------------------------}

function GetCompareFuncName( const CompareFunc : TCompareFunc ) : String;
var
    CompareFuncEntry : TCompareFuncEntry;

begin
    Result := '';

    if GetCompareFuncRegistry.GotoFirst then
    repeat
        CompareFuncEntry := (GetCompareFuncRegistry.CurrentObj as TCompareFuncEntry);
        {$ifdef Win32}
        if @CompareFuncEntry.CompareFunc = @CompareFunc then
        {$else}
        { in Delphi 1.0, the CompareFunc property is giving the compiler
          problems }
        if @CompareFuncEntry.FCompareFunc = @CompareFunc then
        {$endif}
        begin
            Result := CompareFuncEntry.FunctionName;
            Break;
        end;
    until not GetCompareFuncRegistry.GotoNext;

    if Result = '' then
        RaiseCompareFuncNotRegistered;
end;

{--------------------------------------------------------------------------}

function GetCompareFunc( const FunctionName : String ) : TCompareFunc;
var
    Key : TCompareFuncEntry;

begin
    { assign value to result to prevent compiler warning }
    Result := nil;

    Key := TCompareFuncEntry.Create( FunctionName, nil );
    try
        if GetCompareFuncRegistry.Search( Key, 1 ) then
            Result := (GetCompareFuncRegistry.CurrentObj as TCompareFuncEntry).CompareFunc
        else
            RaiseCompareFuncNotRegistered;
    finally
        Key.Free;
    end;
end;

{--------------------------------------------------------------------------}

constructor THashFuncEntry.Create(
                        const InFunctionName : String;
                        const InHashFunc     : THashFunc );
begin
    FFunctionName := InFunctionName;
    FHashFunc     := InHashFunc;

    inherited Create;
end;

{--------------------------------------------------------------------------}

procedure RegisterHashFunction( const FunctionName : String;
                                const HashFunc     : THashFunc );
var
    HashFuncEntry : THashFuncEntry;
begin
    HashFuncEntry := THashFuncEntry.Create( FunctionName, HashFunc );

    if GetHashFuncRegistry.Search( HashFuncEntry, 1 ) then
    begin
        HashFuncEntry.Free;
        RaiseDuplicateHashFunc(FunctionName);
    end
    else
        GetHashFuncRegistry.Insert( HashFuncEntry );
end;

{--------------------------------------------------------------------------}

function GetHashFuncName( const HashFunc : THashFunc ) : String;
var
    HashFuncEntry : THashFuncEntry;

begin
    Result := '';

    if GetHashFuncRegistry.GotoFirst then
    repeat
        HashFuncEntry := (GetHashFuncRegistry.CurrentObj as THashFuncEntry);

        {$ifdef Win32}
        if @HashFuncEntry.HashFunc = @HashFunc then
        {$else}
        { in Dephi 1.0, the CompareFunc property is given the compiler
          problems }
        if @HashFuncEntry.FHashFunc = @HashFunc then
        {$endif}
        begin
            Result := HashFuncEntry.FunctionName;
            Break;
        end;
    until not GetHashFuncRegistry.GotoNext;

    if Result = '' then
        RaiseHashFuncNotRegistered;
end;

{--------------------------------------------------------------------------}

function GetHashFunc( const FunctionName : String ) : THashFunc;
var
    Key : THashFuncEntry;

begin
    { assign value to result to prevent compiler warning }
    Result := nil;

    Key := THashFuncEntry.Create( FunctionName, nil );
    try
        if GetHashFuncRegistry.Search( Key, 1 ) then
            Result := (GetHashFuncRegistry.CurrentObj as THashFuncEntry).HashFunc
        else
            RaiseHashFuncNotRegistered;
    finally
        Key.Free;
    end;
end;

{--------------------------------------------------------------------------}

constructor TObjectStream.Create( const Stream : TStream );
begin
    inherited Create;

    FStream              := Stream;

    FNestedLevel         := 0;

    FFreeStreamOnDestroy := True;

    { MB Oct 27, 1997. Don't invoke Rewind, as it may be desired to read/write
      objects from the current position. }
end;

{--------------------------------------------------------------------------}

constructor TObjectStream.CreateFromFile(
                            const FileName : String;
                            const Mode     : Word);
begin
    FStream := TFileStream.Create( FileName, Mode );

    Self.Create(FStream);
end;

{--------------------------------------------------------------------------}

destructor TObjectStream.Destroy;
begin
    if FreeStreamOnDestroy then
        FStream.Free;

    inherited Destroy;
end;

{--------------------------------------------------------------------------}

function TObjectStream.ReadObject : TObject;
var
    RegisteredClassName : String;
    StreamableInterface : TStreamableInterface;
    ClassVersion        : TClassVersion;

    {$ifdef Ver80}
    ReadProc            : TObjectReadProc;
    {$endif}

begin
    if FNestedLevel = 0 then
        FReader := TObjectReader.Create(Self);

    try
        Inc(FNestedLevel);

        FReader.ReadSignature;

        RegisteredClassName := FReader.ReadString;

        if RegisteredClassName = CNilObjectStrId then
            Result := nil
        else
        begin
            FReader.Read( ClassVersion, SizeOf(ClassVersion) );

            StreamableInterface := GetStreamableInterface( RegisteredClassName );

            Result := StreamableInterface.RegisteredClass.Create;
            try
                {$ifdef Win32}
                StreamableInterface.ReadProcs[ClassVersion]( Result, FReader );
                {$else}
                ReadProc := StreamableInterface.ReadProcs[ClassVersion];
                ReadProc( Result, FReader );
                {$endif}
            except
                Result.Free;
                raise;
            end;
        end;
    finally
        Dec(FNestedLevel);
        if FNestedLevel = 0 then
            FReader.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure TObjectStream.WriteObject( const Obj : TObject );
var
    StreamableInterface : TStreamableInterface;
    ClassVersion        : TClassVersion;

    {$ifdef Ver80}
    WriteProc           : TObjectWriteProc;
    {$endif}

begin
    StreamableInterface := nil; { prevent compiler warning }

    { before attempting to write the object, ensure it is registered by getting
      its StreamableInterface }
    if Obj <> nil then
        StreamableInterface := GetStreamableInterface( Obj.ClassName );

    if FNestedLevel = 0 then
        FWriter := TObjectWriter.Create(Self);
    try
        Inc(FNestedLevel);

        FWriter.WriteSignature;

        if Obj = nil then
            FWriter.WriteString(CNilObjectStrId)
        else
        begin
            FWriter.WriteString( Obj.ClassName );

            ClassVersion := StreamableInterface.DefaultWriteProcVersion;
            FWriter.Write( ClassVersion, SizeOf(ClassVersion));

            {$ifdef Win32}
            StreamableInterface.WriteProcs[ClassVersion]( Obj, FWriter );
            {$else}
            WriteProc := StreamableInterface.WriteProcs[ClassVersion];
            WriteProc( Obj, FWriter );
            {$endif}
        end;
    finally
        Dec(FNestedLevel);
        if FNestedLevel = 0 then
            FWriter.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure TObjectStream.Rewind;
begin
    FStream.Seek(0, soFromBeginning);
end;

{--------------------------------------------------------------------------}

function TObjectStream.GetEndOfStream : Boolean;
begin
    Result := FStream.Position >= FStream.Size;
end;

{--------------------------------------------------------------------------}

constructor TObjectReader.Create( const ObjectStream : TObjectStream );
begin
    FObjectStream := ObjectStream;

    inherited Create( FObjectStream.FStream, CReaderWriterBufSize );
end;

{--------------------------------------------------------------------------}

function TObjectReader.ReadChildObject : TObject;
begin
    Result := FObjectStream.ReadObject;
end;

{--------------------------------------------------------------------------}

constructor TObjectWriter.Create( const ObjectStream : TObjectStream );
begin
    FObjectStream := ObjectStream;

    inherited Create( FObjectStream.FStream, CReaderWriterBufSize );
end;

{--------------------------------------------------------------------------}

procedure TObjectWriter.WriteChildObject( const Obj : TObject );
begin
    FObjectStream.WriteObject(Obj);
end;

{--------------------------------------------------------------------------}

function CloneObj( const Obj : TObject ) : TObject;
var
    MemoryStream : TMemoryStream;

    ObjectStream : TObjectStream;
begin
    MemoryStream := TMemoryStream.Create;

    ObjectStream := TObjectStream.Create(MemoryStream);
    try
        ObjectStream.WriteObject(Obj);

        ObjectStream.Rewind;

        Result := ObjectStream.ReadObject;
    finally
        { the memory stream is freed with the object stream }
        ObjectStream.Free;
    end;
end;

{--------------------------------------------------------------------------}

function StreamsEqual( const StreamA : TStream;
                       const StreamB : TStream ) : Boolean;
var
    ByteA : Byte;
    ByteB : Byte;
begin
    Result := False;

    if StreamA.Size <> StreamB.Size then
        Exit;

    StreamA.Seek(0, soFromBeginning);
    StreamB.Seek(0, soFromBeginning);

    while StreamA.Read(ByteA, SizeOf(ByteA)) = SizeOf(ByteA) do
    begin
        StreamB.Read(ByteB, SizeOf(ByteB));

        if ByteA <> ByteB then
            Exit;
    end;

    Result := True;
end;

{--------------------------------------------------------------------------}

function ObjectsEqual( const ObjA : TObject;
                       const ObjB : TObject ) : Boolean;
var
    StreamA      : TMemoryStream;
    StreamB      : TMemoryStream;
    ObjectStream : TObjectStream;
begin
    StreamA := nil;
    StreamB := nil;
    try
        StreamA      := TMemoryStream.Create;
        ObjectStream := TObjectStream.Create(StreamA);
        try
            ObjectStream.FreeStreamOnDestroy := False;
            ObjectStream.WriteObject(ObjA);
        finally
            ObjectStream.Free;
        end;

        StreamB      := TMemoryStream.Create;
        ObjectStream := TObjectStream.Create(StreamB);
        try
            ObjectStream.FreeStreamOnDestroy := False;
            ObjectStream.WriteObject(ObjB);
        finally
            ObjectStream.Free;
        end;

        Result := StreamsEqual(StreamA, StreamB);

    finally
        StreamB.Free;
        StreamA.Free;
    end;
end;

{--------------------------------------------------------------------------}

function ReadObjectFromFile( const FileName        : String;
                             const UseMemoryStream : Boolean ) : TObject;
var
    ObjectStream : TObjectStream;

    Stream       : TStream;
begin
    if UseMemoryStream then
    begin
        Stream := TMemoryStream.Create;
        try
            TMemoryStream(Stream).LoadFromFile(FileName);
        except
            Stream.Free;
            raise;
        end;
    end
    else
        Stream := TFileStream.Create( FileName, fmOpenRead );

    ObjectStream  := TObjectStream.Create(Stream);
    try
        Result := ObjectStream.ReadObject;
    finally
        ObjectStream.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure WriteObjectToFile( const Obj             : TObject;
                             const FileName        : String;
                             const UseMemoryStream : Boolean );
var
    ObjectStream : TObjectStream;

    Stream       : TStream;
begin
    if UseMemoryStream then
        Stream := TMemoryStream.Create
    else
        Stream := TFileStream.Create( FileName, fmCreate );

    ObjectStream := TObjectStream.Create(Stream);
    try
        ObjectStream.WriteObject(Obj);

        if UseMemoryStream then
            TMemoryStream(Stream).SaveToFile(FileName);
    finally
        ObjectStream.Free;
    end;
end;

{--------------------------------------------------------------------------}

type

TProtectedGraphic = class(TGraphic);

procedure WriteGraphic( const Obj     : TObject;
                        const Writer  : TObjectWriter );
var
    Graphic : TProtectedGraphic;
    Stream  : TMemoryStream;
begin
    Graphic := TProtectedGraphic(Obj as TGraphic);
    Stream := TMemoryStream.Create;
    try
        Graphic.WriteData(Stream);
        Writer.WriteInteger(Stream.Size);
        Writer.Write(Stream.Memory^, Stream.Size);
    finally
        Stream.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure ReadGraphic( const Obj     : TObject;
                       const Reader  : TObjectReader );
var
    Graphic : TProtectedGraphic;
    Stream  : TMemoryStream;
begin
    Graphic := TProtectedGraphic(Obj as TGraphic);
    Stream := TMemoryStream.Create;
    try
        Stream.SetSize(Reader.ReadInteger);
        Reader.Read(Stream.Memory^, Stream.Size);
        Graphic.ReadData(Stream);
    finally
        Stream.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure WriteMemoryStream( const Obj     : TObject;
                             const Writer  : TObjectWriter );
var
    Stream  : TMemoryStream;
begin
    Stream := Obj as TMemoryStream;
    Stream.Seek(0, soFromBeginning);

    Writer.WriteInteger(Stream.Size);
    Writer.Write(Stream.Memory^, Stream.Size);
end;
{--------------------------------------------------------------------------}

procedure ReadMemoryStream( const Obj     : TObject;
                            const Reader  : TObjectReader );
var
    Stream : TMemoryStream;
    Size   : LongInt;
begin
    Stream := Obj as TMemoryStream;

    Size   := Reader.ReadInteger;
    Stream.SetSize(Size);
    Reader.Read(Stream.Memory^, Size);
end;

{--------------------------------------------------------------------------}

initialization

    { Ensure the registry objects are created. }
    GetObjectRegistry;
    GetCompareFuncRegistry;
    GetHashFuncRegistry;

    {$ifdef Win32}
    RegisterStreamable( TBitmap,
                        [ TObjectReadProc(@ReadGraphic) ],
                        [ TObjectWriteProc(@WriteGraphic) ]
                       );
    RegisterStreamable( TIcon,
                        [ TObjectReadProc(@ReadGraphic) ],
                        [ TObjectWriteProc(@WriteGraphic) ]
                       );
    RegisterStreamable( TMetaFile,
                        [ TObjectReadProc(@ReadGraphic) ],
                        [ TObjectWriteProc(@WriteGraphic) ]
                       );
    RegisterStreamable( TMemoryStream,
                        [ TObjectReadProc(@ReadMemoryStream) ],
                        [ TObjectWriteProc(@WriteMemoryStream) ]
                       );
    {$else}
    RegisterStreamable( TBitmap,
                        [ReadGraphic],
                        [WriteGraphic]
                      );
    RegisterStreamable( TIcon,
                        [ReadGraphic],
                        [WriteGraphic]
                      );
    RegisterStreamable( TMetaFile,
                        [ReadGraphic],
                        [WriteGraphic]
                      );
    RegisterStreamable( TMemoryStream,
                        [ ReadMemoryStream ],
                        [ WriteMemoryStream ]
                       );
    {$endif}

    InitializeCriticalSection(GetStreamableInterfaceCS);

finalization

    DeleteCriticalSection(GetStreamableInterfaceCS);

    HashFuncRegistry.Free;
    CompareFuncRegistry.Free;
    ObjectRegistry.Free;
end.
