{----------------------------------------------------------------------
|
| Library: Spider Object Database
|
| Module: OdbBtrFl.Pas
|
| Description: Btrieve file class used with Btrieve object database driver
|              class.
|
| History: Jan 25, 1998. Michel Brazeau, first version
|
|---------------------------------------------------------------------}
unit OdbBtrFl;

{$I Odb.Inc}

interface

uses
    Classes,  { for TMemoryStream }
    SysUtils, { for Exception }
    ObjList,  { for TUnOrderedList }
    Odb;      { for TObjectFile }

const

{ btrieve constants used by TBtrieveFile. There are many more which may be
  used by the Btrieve API. }
bMaxDatabufSize      = 57000;
bMaxKeySeg           = 119;
bPosBlockSize        = 128;
bMaxKeySize          = 255;
bMaxPageSize         = 4096;
bMaxStatbufSize      = 33455;

bXtractrDirectSign   = $80000000;
bTruncSign           = $80000004;

bNormal              = $00;
bExclusive           = $FC;

bVarRecs             = $0001;

bNoError             = 0;
bKeyNotFound         = 4;
bEOF                 = 9;
bDatalengthError     = 22;
bPositionNotSet      = 8;


bDuplicate           = $0001;
bModifiable          = $0002;
bSegmented           = $0010;
bDescendingKey       = $0040;
bExtTypeKey          = $0100;
bNocaseKey           = $0400;

bOpen                = 0;
bClose               = 1;
bInsert              = 2;
bUpdate              = 3;
bDelete              = 4;
bGetEqual            = 5;
bGetNext             = 6;
bGetPrev             = 7;
bGetGreatEqual       = 9;
bGetFirst            = 12;
bGetLast             = 13;
bCreate              = 14;
bStat                = 15;
bBeginTransaction    = 19;
bEndTransaction      = 20;
bAbortTransaction    = 21;
bGetPosition         = 22;
bGetDirect           = 23;
bStop                = 25;
bVersion             = 26;
bUnlock              = 27;
bChunkUpdate         = 53;

bSingleNoWaitLock    = 200;
bMultiNoWaitLock     = 400;
bConCurrentTransBias = 1000;

bStringKey           = 0;
bIntegerKey          = 1;
bIEEEKey             = 2;
bLStringKey          = 10;
bAutoincrementKey    = 15;

type

{--------------------------------------------------------------------------}

{ data buffer }
TpBtDataBuffer = ^TBtDataBuffer;
TBtDataBuffer = array[0..bMaxDatabufSize-1] of Byte;

{ file specification }
TpBtFileSpec = ^TBtFileSpec;
TBtFileSpec = packed record
    RecLen      : Word;
    PageSize    : Word;
    NoIndices   : Word;
    NoRecords   : LongInt;
    FileFlags   : Word;
    Reserved    : Word;
    Allocation  : Word;
end;

{--------------------------------------------------------------------------}

{ key segment specification }
TpBtKeySpec = ^TBtKeySpec;
TBtKeySpec = packed record
    KeyPos        : Word;          { 1 based }
    KeyLen        : Word;
    KeyFlags      : Word;
    NoKeys        : LongInt;
    ExtKeyType    : Byte;
    NullValue     : Byte;
    Reserved      : Word;
    KeyNo         : Byte;
    ACSNo         : Byte;
end;

{--------------------------------------------------------------------------}

{ array of key segments }

TpBtKeySpecArray = ^TBtKeySpecArray;
TBtKeySpecArray = array[0..(bMaxKeySeg-1)] of TBtKeySpec;

{--------------------------------------------------------------------------}

{ position block }

TpBtPositionRec = ^TBtPositionRec;
TBtPositionRec = packed record
    Block : array[1..bPosBlockSize] of Byte;
end;

{ key buffer }
TpBtKeyBuffer = ^TBtKeyBuffer;
TBtKeyBuffer = array[1..bMaxKeySize] of Byte;

{--------------------------------------------------------------------------}

{ client block }

TpBtClientId = ^TBtClientId;
TBtClientId = packed record
    Reserved : array[1..12] of Byte;
    {$ifdef Win32}
    InstanceId : array[1..2] of AnsiChar;
    {$else}
    InstanceId : array[1..2] of Char;
    {$endif}
    TaskId     : Word;
end;

{--------------------------------------------------------------------------}

EBtrInvalidClientId           = class(Exception);
EBtrieveEngineNotInstalled    = class(Exception);
EBtrieveFileStructureMismatch = class(Exception);

{--------------------------------------------------------------------------}

procedure NewClientId( var ClientId : TBtClientId );

{--------------------------------------------------------------------------}

function BtrvId(     Operation      : Word;
                 var PositionBlock;
                 var DataBuffer;
                 var DataLen        : Word;
                 var KeyBuffer;
                     KeyNumber      : SmallInt;
                 var ClientId
               ) : SmallInt;


const

CBtrieveFileExtension = '.Btr';

{--------------------------------------------------------------------------}

type

TBtrieveFile = class(TObject)
protected
    { related object file. }
    FObjectFile   : TObjectFile;

    FFileSpec     : TBtFileSpec;

    FKeySpecCount : Integer; { Number of key specs  in FKeySpecs. }
    FKeySpecs     : TBtKeySpecArray;

    FPositionRec  : TBtPositionRec;
    FKeyBuffer    : TBtKeyBuffer;

    { Current Btrieve index number which is 0 based. }
    FIndexNo      : Integer;

    FDiskFileName : String[255];

    pDataBuffer   : TpBtDataBuffer;

    pClientId     : TpBtClientId;

    function GetIndexName : String;
    procedure SetIndexName( const InIndexName : String );

    { update the file specification from FObjectFile }
    procedure UpdateFileSpec;

    { update the key specifications from FObjectFile }
    procedure UpdateKeySpecs;

    { Updates FFileSpec and FKeySpecs according to database file
      structure. If the physical structure does not match the current one,
      EBtrieveFileStructureMismatch is raised. If InVerifyIndices is False,
      the index structure of the physical file is not verified nor read. }
    procedure ReadAndCompareStats( const VerifyIndices : Boolean );

    { Updates FKeyBuffer according to the current data in pDataBuffer^ and
      the current index specifed by FIndexNo. }
    procedure UpdateKeyBuffer;

    { Transfers the field values in Fields to the data buffer. If
      ClearAutoInc is True, autoincrement fields in the buffer are set
      to 0. }
    procedure TransferFieldsToBuffer( const Fields       : TUnOrderedList;
                                      const ClearAutoInc : Boolean );

    { Transfer the autoincrement fields in pDataBuffer^ to the appropriate
      fields  in Fields. }
    procedure TransferAutoIncFieldsToBuffer( const Fields : TUnOrderedList );

    { Get the ObjectSize stored in the data buffer. }
    function GetObjectSize : LongInt;

    { Set the ObjectSize stored in the data buffer. }
    procedure SetObjectSize( const ObjectSize : LongInt );

    property ObjectSize : LongInt read GetObjectSize
                                  write SetObjectSize;

    { Insert the object data contained in Stream for the current record
      positioned. }
    procedure WriteObjectData( const Stream  : TMemoryStream );

public
    constructor Create( const InObjectFile : TObjectFile );
    destructor Destroy; override;

    procedure CreateFile;
    procedure DeleteFile;

    procedure Open;
    procedure OpenExclusive;
    procedure Close;

    procedure Insert( const Fields  : TUnOrderedList;
                      const Stream  : TMemoryStream );

    procedure Update( const Fields  : TUnOrderedList;
                      const Stream  : TMemoryStream );

    procedure Delete;

    property IndexName : String read GetIndexName
                                write SetIndexName;

    function LockCurrent : Boolean;
    procedure UnLockCurrent;

    function First : Boolean;
    function Last : Boolean;
    function Next : Boolean;
    function Prior : Boolean;

    function FindExact( const Fields : TUnOrderedList ) : Boolean;
    procedure FindNearest( const Fields : TUnOrderedList );

    function GetBookmark : Pointer;
    procedure GotoBookmark( const Bookmark : Pointer );
    procedure FreeBookmark( const Bookmark : Pointer );

    function RecordCount : LongInt;

    { Read the object data belonging to the current record. }
    procedure ReadObjectData( const Stream  : TMemoryStream );

    { Returns the physical position of the current record. Returns -1, if
      position has not been established. }
    function GetPosition : LongInt;

    { Positions the current record given a physical position. If Position
      is -1, False is returned without raising an exception. }
    function GetDirect( const Position : LongInt ) : Boolean;

end;

{--------------------------------------------------------------------------}

EOdbBtrieveError    = class(Exception);

TBtrieveStatusFunc = function( const Status : Integer ) : String;

{--------------------------------------------------------------------------}

procedure BtrieveCheck( const Status   : Integer;
                        const FileName : String );

{--------------------------------------------------------------------------}

var

{ Global function pointer which returns a status message corresponding to a
Btrieve status code. May be changed for multilingual support. }
BtrieveStatusFunc : TBtrieveStatusFunc;

{--------------------------------------------------------------------}

implementation

uses
    {$ifdef Win32}
    Windows,  { GetTickCount }
    {$else}
    WinTypes,
    WinProcs,
    {$endif}
    ObjBuckt, { for TStringCombo }
    OdbIndex, { for TOdbIndex }
    OdbField, { for TOdbField }
    OdbMisc,  { for RaiseOdbException,  GetLocalObjectFilename }
    OdbBtr;   { for TBtrieveOdbDriver }

{--------------------------------------------------------------------------}

const

CLockTries            = 5;
CBetweenLockDelayMS   = 200;

{--------------------------------------------------------------------------}

type

{ Btrieve bookmark. }
TpBtrieveBookmark = ^TBtrieveBookmark;
TBtrieveBookmark = packed record
    Position  : LongInt;
    IndexNo   : Integer;
end;

{--------------------------------------------------------------------------}

TpChunkRec = ^TChunkRec;
TChunkRec  = packed record
    SubFunction : LongInt;
    ChunkCount  : LongInt;
    Offset      : LongInt;
    Length      : LongInt;
    UserData    : LongInt;
end;

{--------------------------------------------------------------------------}

TpTruncateChunkRec = ^TTruncateChunkRec;
TTruncateChunkRec  = packed record
    SubFunction : LongInt;
    Offset      : LongInt;
end;

{--------------------------------------------------------------------------}

{ Class to facilitate the building of btrieve key specifications. }
TBtKeySpecBuilder = class(TObject)
public

    constructor Create( var KeySpecArray : TBtKeySpecArray );

    destructor Destroy; override;


    { Adds a new index. Has to be followed by one or more calls
      to AddSegment to define the fields in the index. May contain
      bModifiable, bDuplicate }
    procedure AddIndex( IndexFlags : Word );

    { Adds a field to the current index. StartPos is the position of
    the field. Length is the length in bytes of the field. SegmentFlags are
    Btrieve segment flags, (bDescending, bNocaseKey). ExtFlag is the Btrieve
    extended key type (bStringKey, bIntegerKey, bIEEEKey) }
    procedure AddField( StartPos     : Word;
                        Length       : Word;
                        SegmentFlags : Word;
                        ExtFlag      : Word
                      );


private

    pKeySpecArray   : TpBtKeySpecArray;

    { A call to add segment, adds a segment at the CurSegment index, ie
    CurSegment is Post incremented. CurSegment is 0 based. }
    CurSegment      : Word;

    { The Btrieve flags for the current index. }
    CurIndexFlags   : Word;

    { True on the first call to AddSegment, after a call to AddIndex. }
    FirstSegment    : Boolean;

    { Btrieve key number for the current segment, automatically incremented
    for each new index. }
    CurIndexNo      : Integer;
end;

{--------------------------------------------------------------------------}

constructor TBtKeySpecBuilder.Create( var KeySpecArray   : TBtKeySpecArray );
begin
    inherited Create;

    pKeySpecArray   := @KeySpecArray;

    CurSegment      := 0;
    CurIndexNo      := -1; { must start at -1, incrmented on each call to
                             add index }

    { clear the key spec array }
    FillChar(pKeySpecArray^, SizeOf(TBtKeySpecArray), 0);
end;

{------------------------------------------------------------------------}

destructor TBtKeySpecBuilder.Destroy;
begin
    inherited Destroy;
end;

{------------------------------------------------------------------------}

procedure TBtKeySpecBuilder.AddIndex( IndexFlags : Word );
begin
    FirstSegment  := True;
    CurIndexFlags := IndexFlags;

    Inc(CurIndexNo);
end;

{------------------------------------------------------------------------}

procedure TBtKeySpecBuilder.AddField( StartPos     : Word;
                                      Length       : Word;
                                      SegmentFlags : Word;
                                      ExtFlag      : Word
                                    );
begin
    if CurSegment >= High(TBtKeySpecArray) then
        Exit;

    { if the previous segment is part of the same index, mark it as
      bSegmented.  }
    if (not FirstSegment) and (CurSegment > 0) then
        SetWordFlags(Word(pKeySpecArray^[CurSegment-1].KeyFlags), bSegmented);

    with pKeySpecArray^[CurSegment] do
    begin
        KeyPos     := StartPos + 1; { convert from 0 based to 1 based }
        KeyLen     := Length;
        KeyFlags   := CurIndexFlags or SegmentFlags;
        KeyNo      := CurIndexNo;

        if ExtFlag <> 0 then
        begin
            KeyFlags   := KeyFlags or bExtTypeKey;
            ExtKeyType := ExtFlag;
        end;
    end;

    Inc(CurSegment);

    FirstSegment := False;

end;

{------------------------------------------------------------------------}

procedure NewClientId( var ClientId : TBtClientId );
begin
    FillChar(ClientId, SizeOf(ClientId), 0);

    { The ClientId should be choosen randomly to avoid to database sessions
      having the same ClientId. }
    Randomize;
    ClientId.InstanceId[1] := Chr( Ord('A') + Random(Ord('z') - Ord('A')));
    ClientId.InstanceId[2] := Chr( Ord('A') + Random(Ord('z') - Ord('A')));

    { Attempt to make TaskId unique each ClientId by using the ellapsed milli
      seconds since Windows has started. }
    RandSeed := GetTickCount;
    ClientId.TaskId := Random($FFFF);
end;

{--------------------------------------------------------------------------}

function BtrCallId(     Operation   : Word;
                    var PosBlk;
                    var DataDuf;
                    var DataLen     : LongInt;
                    var KeyBuf;
                        KeyLen      : Byte;
                        KeyNum      : Byte;
                    var ClientId
                  ) : SmallInt;
                 {$ifdef Win32}
                 stdcall; external 'WBTRV32.DLL'  name 'BTRCALLID';
                 {$else}
                 far; external 'WBTRCALL.DLL'  name 'BTRCALLID';
                 {$endif}

{--------------------------------------------------------------------------}

function BtrvId(     Operation      : Word;
                 var PositionBlock;
                 var DataBuffer;
                 var DataLen        : Word;
                 var KeyBuffer;
                     KeyNumber      : SmallInt;
                 var ClientId
               ) : SmallInt;
type
    TWordArray = array[1..2] of Word;
var
    KeyLen      : Byte;
    TempDataLen : LongInt;
begin
    KeyLen      := 255;
    TempDataLen := DataLen;
    BtrvId := BtrCallId( Operation,
                         PositionBlock,
                         DataBuffer,
                         TempDataLen,
                         KeyBuffer,
                         KeyLen,
                         KeyNumber,
                         ClientId );

    DataLen := TWordArray(TempDataLen)[1];
end;

{--------------------------------------------------------------------------}

procedure BtrieveCheck( const Status   : Integer;
                        const FileName : String );
begin
    { Filter out "normal" status codes. }
    if (Status <> bNoError) and
       (Status <> bKeyNotFound) and
       (Status <> bEOF) then
        RaiseOdbException( EOdbBtrieveError, '#' + IntToStr(Status) + '. ' +
                           BtrieveStatusFunc(Status) + '. ' + #13 +
                           FileName);
end;

{--------------------------------------------------------------------------}

constructor TBtrieveFile.Create( const InObjectFile : TObjectFile );
begin
    inherited Create;

    FObjectFile := InObjectFile;

    UpdateFileSpec;
    UpdateKeySpecs;

    FillChar(FPositionRec, SizeOf(FPositionRec), 0);

    FDiskFileName :=  GetLocalObjectFileName( FObjectFile.Database.DatabaseName,
                                              FObjectFile.ObjectFileName,
                                              CBtrieveFileExtension );

    { Null terminate the file name for Btrieve API calls. }
    if Length(FDiskFilename) < (SizeOf(FDiskFileName)-1) then
        FDiskFileName[Length(FDiskFilename)+1] := #0
    else
        FDiskFileName[SizeOf(FDiskFileName)-1] := #0;

    pClientId := (FObjectFile.Driver as TBtrieveOdbDriver).GetClientIdPtr;

    GetMem(pDataBuffer, FFileSpec.RecLen);
end;

{--------------------------------------------------------------------------}

destructor TBtrieveFile.Destroy;
begin
    if pDataBuffer <> nil then
        FreeMem(pDataBuffer, FFileSpec.RecLen);

    inherited Destroy;
end;

{--------------------------------------------------------------------------}

{ Returns the size of a Btrieve field given a TOdbField. }
function BtrieveFieldSize( const Field : TOdbField ) : LongInt;
begin
    { prevent compiler warning. }
    Result := 0;

    case Field.FieldType of
    oftInteger,
    oftAutoInc   : Result := SizeOf(LongInt);
    oftFloat     : Result := SizeOf(Double);
    oftString    : Result := Field.Size;
    else
        RaiseOdbException( EOdbError, 'Unknown field type');
    end;
end;

{--------------------------------------------------------------------------}

{ Returns the Btrieve extended field flag given a TOdbField. }
function BtrieveFieldType( const Field : TOdbField ) : Word;
begin
    { prevent compiler warning. }
    Result := 0;

    case Field.FieldType of
    oftInteger  : Result := bIntegerKey;
    oftAutoInc  : Result := bAutoIncrementKey;
    oftFloat    : Result := bIEEEKey;
    oftString   : Result := bLStringKey;
    else
        RaiseOdbException(EOdbError, 'Unknown field type');
    end;
end;

{--------------------------------------------------------------------------}

{ Returns the offset of a field. }
function FieldOffset( const FieldName : String;
                      const Fields    : TUnOrderedList ) : LongInt;
var
    Field         : TOdbField;
    Found         : Boolean;
begin
    if Fields.Size = 0 then
        RaiseOdbException(EOdbError, 'Field list is empty');

    Found  := False;
    Result := 0;
    if Fields.GotoFirst then
    repeat
        Field := Fields.CurrentObj as TOdbField;
        if Field.Name = FieldName then
        begin
            Found := True;
            Break;
        end;

        Inc(Result, BtrieveFieldSize(Field));
    until not Fields.GotoNext;

    if not Found then
        RaiseOdbException(EOdbError, 'Field not found');
end;

{--------------------------------------------------------------------------}

type

Str255 = String[255];

procedure TBtrieveFile.TransferFieldsToBuffer(
                                const Fields       : TUnOrderedList;
                                const ClearAutoInc : Boolean );
var
    Field     : TOdbField;
    Offset    : LongInt;
    pValue    : Pointer;
begin
    Offset := 0;
    if Fields.GotoFirst then
    repeat
        Field  := Fields.CurrentObj as TOdbField;
        pValue := @pDataBuffer^[Offset];

        case Field.FieldType of
            oftInteger : LongInt(pValue^) := Field.AsInteger;
            oftAutoInc : if ClearAutoInc then
                             LongInt(pValue^) := 0
                         else
                             LongInt(pValue^) := Field.AsInteger;
            oftString  : Str255(pValue^)  := Field.AsString;
            oftFloat   : Double(pValue^)  := Field.AsFloat;
        end;
        Inc(Offset, BtrieveFieldSize(Field));
    until not Fields.GotoNext;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.TransferAutoIncFieldsToBuffer(
                                   const Fields : TUnOrderedList );
var
    Field     : TOdbField;
    Offset    : LongInt;
    pValue    : Pointer;
begin
    Offset := 0;
    if Fields.GotoFirst then
    repeat
        Field  := Fields.CurrentObj as TOdbField;
        pValue := @pDataBuffer^[Offset];

        case Field.FieldType of
            oftAutoInc : Field.AsInteger := LongInt(pValue^);
        end;
        Inc(Offset, BtrieveFieldSize(Field));
    until not Fields.GotoNext;
end;

{--------------------------------------------------------------------------}

function TBtrieveFile.GetObjectSize : LongInt;
var
    pValue : Pointer;
begin
    pValue := @pDataBuffer^[FFileSpec.RecLen-4];
    Result := LongInt(pValue^);
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.SetObjectSize( const ObjectSize : LongInt );
var
    pValue : Pointer;
begin
    pValue  := @pDataBuffer^[FFileSpec.RecLen-4];
    LongInt(pValue^) := ObjectSize;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.UpdateFileSpec;

    function ComputeRecordLength : Word;
    var
        Field : TOdbField;
    begin
        Result := 0;
        if FObjectFile.Fields.GotoFirst then
        repeat
             Field := FObjectFile.Fields.CurrentObj as TOdbField;
             Result := Result + BtrieveFieldSize(Field);
        until not FObjectFile.Fields.GotoNext;

        { Account for the ObjectSize internal field. }
        Result := Result + 4;
    end;

begin
    FillChar(FFileSpec, SizeOf(FFileSpec), 0);

    with FFileSpec do
    begin
        RecLen     := ComputeRecordLength;
        PageSize   := bMaxPageSize;
        NoIndices  := FObjectFile.Indices.Size;
        FileFlags  := bVarRecs;
    end;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.UpdateKeySpecs;
var
    KeySpecBuilder : TBtKeySpecBuilder;
    Index          : TOdbIndex;
    Field          : TOdbField;
    IndexFlags     : Word;
    SegmentFlags   : Word;
    FieldNo        : Integer;

begin
    FKeySpecCount  := 0;

    KeySpecBuilder := TBtKeySpecBuilder.Create(FKeySpecs);
    try
        if FObjectFile.Indices.GotoFirst then
        repeat
            Index := FObjectFile.Indices.CurrentObj as TOdbIndex;
            IndexFlags := bModifiable;
            if not (iaPrimary in Index.Attributes) and
               not (iaUnique in Index.Attributes) then
                IndexFlags := IndexFlags or bDuplicate;

            KeySpecBuilder.AddIndex(IndexFlags);

            for FieldNo := 1 to Index.FieldNames.Size do
            begin
                Inc(FKeySpecCount);

                SegmentFlags := 0;
                if iaDescending in Index.Attributes then
                    SegmentFlags := bDescendingKey;

                Field := FObjectFile.FieldByName(
                             (Index.FieldNames[FieldNo] as TStringCombo).Str);

                if (Field.FieldType = oftString) and
                   (iaCaseInsensitive in Index.Attributes) then
                    SegmentFlags := SegmentFlags or bNoCaseKey;

                KeySpecBuilder.AddField( FieldOffset( Field.Name,
                                                      FObjectFile.Fields),
                                         BtrieveFieldSize(Field),
                                         SegmentFlags,
                                         BtrieveFieldType(Field) );
            end;

        until not FObjectFile.Indices.GotoNext;
    finally
        KeySpecBuilder.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.ReadAndCompareStats( const VerifyIndices : Boolean );
var
    CurSegment    : Word;
    CurIndex      : Word;
    pBuffer       : TpBtDataBuffer;
    pKeyBuffer    : TpBtKeyBuffer;
    BufferLen     : Word;
    pFileSpec     : TpBtFileSpec;
    pKeySpec      : TpBtKeySpec;

begin
    { MB Jun 03, 1998. Use a temporay key buffer, otherwise position will
      be lost. } 
    GetMem(pKeyBuffer, SizeOf(TBtKeyBuffer));
    GetMem(pBuffer, bMaxStatbufSize);
    try
        FillChar(pBuffer^, bMaxStatbufSize, 0);
        BufferLen := bMaxStatbufSize;

        { the BufSize variable will be changed in the call to Btrv, so it
          is not used in memory allocation/deallocation }
        BtrieveCheck( BtrvId( bStat, FPositionRec, pBuffer^, BufferLen,
                              pKeyBuffer^, 0, pClientId^),
                      FDiskFileName );

        { read and compare the file specs }
        pFileSpec := Pointer(pBuffer);

        { MB Apr 06, 2001. Only compare RecLen as Pervasive.SQL has new
          FileFlags that failed the comparison }
        if (pFileSpec^.RecLen    <> FFileSpec.RecLen) or
           (
             (pFileSpec^.NoIndices  <> FFileSpec.NoIndices) and
             VerifyIndices
           ) then
            RaiseOdbException(EBtrieveFileStructureMismatch, FDiskFileName);

        Move(pFileSpec^, FFileSpec, SizeOf(FFileSpec));

        if VerifyIndices then
        begin
            CurSegment := 0;
            CurIndex   := 1;

            while CurIndex <= FFileSpec.NoIndices do
            begin
                { read and compare key spec }
                pKeySpec := Pointer(@pBuffer^[ SizeOf(FFileSpec)+
                                               (CurSegment*SizeOf(TBtKeySpec))]);

                if (pKeySpec^.KeyPos     <> FKeySpecs[CurSegment].KeyPos) or
                   (pKeySpec^.KeyLen     <> FKeySpecs[CurSegment].KeyLen) or
                   (pKeySpec^.KeyNo      <> FKeySpecs[CurSegment].KeyNo) or
                   (pKeySpec^.ExtKeyType <> FKeySpecs[CurSegment].ExtKeyType) then
                    RaiseOdbException(EBtrieveFileStructureMismatch, FDiskFileName);

                Move(pKeySpec^, FKeySpecs[CurSegment], SizeOf(TBtKeySpec));

                if not AreFlagsSet(FKeySpecs[CurSegment].KeyFlags, bSegmented) then
                    Inc(CurIndex);

                Inc(CurSegment);

            end; { while }
        end; { not IgnoreIndexes }
    finally
        FreeMem(pBuffer, bMaxStatbufSize);
        FreeMem(pKeyBuffer, SizeOf(TBtKeyBuffer));
    end;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.Open;
var
    DataLen : Word;
begin
    DataLen := 0;
    BtrieveCheck( BtrvId( bOpen, FPositionRec, pDataBuffer^, DataLen,
                          FDiskFileName[1], bNormal,
                          pClientId^ ),
                  FDiskFileName );

    ReadAndCompareStats(True);
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.OpenExclusive;
var
    DataLen : Word;
begin
    DataLen := 0;
    BtrieveCheck( BtrvId( bOpen, FPositionRec, pDataBuffer^, DataLen,
                          FDiskFileName[1], bExclusive,
                          pClientId^ ),
                  FDiskFileName );

    ReadAndCompareStats(True);
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.Close;
var
    DataLen : Word;
begin
    DataLen := 0;
    BtrieveCheck( BtrvId( bClose, FPositionRec, pDataBuffer^, DataLen,
                          FDiskFileName[1], 0, pClientId^ ),
                  FDiskFileName );
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.CreateFile;
var
    BufferLen : Word;
    pBuffer   : TpBtDataBuffer;
    Index     : Integer;
begin
    BufferLen := SizeOf(TBtFileSpec) + FKeySpecCount*SizeOf(TBtKeySpec);
    GetMem(pBuffer, BufferLen);
    try
        FillChar(pBuffer^, BufferLen, 0);

        Move(FFileSpec, pBuffer^[0], SizeOf(FFileSpec));

        for Index := 0 to (FKeySpecCount-1) do
        begin
            Move( FKeySpecs[Index],
                  pBuffer^[SizeOf(FFileSpec)+(Index*SizeOf(TBtKeySpec))],
                  SizeOf(TBtKeySpec));
        end;

        BtrieveCheck( BtrvId( bCreate, FPositionRec, pBuffer^, BufferLen,
                              FDiskFileName[1], 0, pClientId^ ),
                      FDiskFileName );
    finally
        FreeMem(pBuffer, BufferLen);
    end;

end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.DeleteFile;
begin
    if FObjectFile.Opened then
        FObjectFile.Close;

    SysUtils.DeleteFile(FDiskFileName);
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.WriteObjectData( const Stream       : TMemoryStream );
var
    BytesRead         : LongInt;
    ChunkSize         : LongInt;
    pBuffer           : TpBtDataBuffer;
    BufferSize        : LongInt;
    pChunkRec         : TpChunkRec;
    Offset            : LongInt;
    DataLen           : Word;

begin
    ChunkSize  := (FObjectFile.Driver as TBtrieveOdbDriver).ChunkSize;
    BufferSize := SizeOf(TChunkRec) + ChunkSize;

    GetMem(pBuffer, BufferSize);
    try
        Offset := FFileSpec.RecLen;

        Stream.Seek(0, soFromBeginning);

        repeat
            pChunkRec := TpChunkRec(pBuffer);

            BytesRead := Stream.Read( pBuffer^[SizeOf(TChunkRec)],
                                      ChunkSize );

            if BytesRead > 0 then
            begin
                pChunkRec^.SubFunction := LongInt(bXtractrDirectSign);
                pChunkRec^.ChunkCount  := 1;
                pChunkRec^.Offset      := Offset;
                pChunkRec^.Length      := BytesRead;
                pChunkRec^.UserData    := 0;

                DataLen := SizeOf(TChunkRec) + BytesRead;

                BtrieveCheck( BtrvId( bChunkUpdate, FPositionRec,
                                      pBuffer^, DataLen,
                                      FKeyBuffer, FIndexNo, pClientId^),
                              FDiskFileName );

                Inc(Offset, BytesRead);
            end;
        until BytesRead < ChunkSize;

    finally
        FreeMem(pBuffer, BufferSize);
    end;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.ReadObjectData( const Stream  : TMemoryStream );
var
    BytesLeft         : LongInt;
    ChunkSize         : LongInt;
    pBuffer           : TpBtDataBuffer;
    BufferSize        : LongInt;
    pChunkRec         : TpChunkRec;
    DataLen           : Word;
    Offset            : LongInt;
    Position          : LongInt;

begin
    ChunkSize  := (FObjectFile.Driver as TBtrieveOdbDriver).ChunkSize;

    Position   := GetPosition;

    if Position = -1 then
        BtrieveCheck( bPositionNotSet, FDiskFileName );

    BufferSize := ChunkSize;
    if BufferSize < (SizeOf(TChunkRec) + 4) then
        BufferSize := (SizeOf(TChunkRec) + 4);

    GetMem(pBuffer, BufferSize);
    try
        Stream.Clear;

        BytesLeft := ObjectSize;
        Offset    := FFileSpec.RecLen;

        while BytesLeft > 0 do
        begin
            LongInt(Pointer(pBuffer)^) := Position;
            pChunkRec := TpChunkRec(@pBuffer^[4]);

            pChunkRec^.SubFunction := LongInt(bXtractrDirectSign);
            pChunkRec^.ChunkCount  := 1;
            pChunkRec^.Offset      := Offset;
            if BytesLeft < ChunkSize then
                pChunkRec^.Length := BytesLeft
            else
                pChunkRec^.Length := ChunkSize;
            pChunkRec^.UserData    := 0;

            DataLen := BufferSize;

            { Decrement BytesLeft before the Btrieve call as the data buffer
              will be modified. }
            Dec(BytesLeft, pChunkRec^.Length);
            Inc(Offset, pChunkRec^.Length);

            BtrieveCheck( BtrvId( bGetDirect, FPositionRec,
                                  pBuffer^, DataLen,
                                  FKeyBuffer, $FE, pClientId^),
                          FDiskFileName );

            Stream.Write(pBuffer^, DataLen);
        end;

    finally
        FreeMem(pBuffer, BufferSize);
    end;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.Insert( const Fields  : TUnOrderedList;
                               const Stream  : TMemoryStream );
var
    DataLen : Word;
begin
    FillChar(pDataBuffer^, SizeOf(FFileSpec.RecLen), 0);

    TransferFieldsToBuffer(Fields, True);
    ObjectSize := Stream.Size;

    DataLen := FFileSpec.RecLen;

    BtrieveCheck( BtrvId( bInsert, FPositionRec, pDataBuffer^, DataLen,
                          FKeyBuffer, FIndexNo, pClientId^),
                  FDiskFileName );

    WriteObjectData(Stream);

    { Update any autoincrement field in Fields. }
    TransferAutoIncFieldsToBuffer(Fields);
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.Update( const Fields  : TUnOrderedList;
                               const Stream  : TMemoryStream );
var
    pBuffer           : TpBtDataBuffer;
    BufferSize        : LongInt;
    pChunkRec         : TpChunkRec;
    pTruncateChunkRec : TpTruncateChunkRec;
    DataLen           : Word;
begin
    BufferSize := SizeOf(TChunkRec) + FFileSpec.RecLen;

    GetMem(pBuffer, BufferSize);
    try
        if Stream.Size < ObjectSize then
        begin
            { If this is an update, remove the previous object data.}
            pTruncateChunkRec := TpTruncateChunkRec(pBuffer);

            pTruncateChunkRec^.SubFunction := LongInt(bTruncSign);
            pTruncateChunkRec^.Offset      := FFileSpec.RecLen + Stream.Size;

            DataLen := SizeOf(TTruncateChunkRec);

            BtrieveCheck( BtrvId( bChunkUpdate, FPositionRec,
                                  pBuffer^, DataLen,
                                  FKeyBuffer, FIndexNo, pClientId^),
                          FDiskFileName );
        end;

        TransferFieldsToBuffer(Fields, False);
        ObjectSize := Stream.Size;

        { Update the fixed portion. }

        pChunkRec := TpChunkRec(pBuffer);

        pChunkRec^.SubFunction := LongInt(bXtractrDirectSign);
        pChunkRec^.ChunkCount  := 1;
        pChunkRec^.Offset      := 0;
        pChunkRec^.Length      := FFileSpec.RecLen;
        pChunkRec^.UserData    := 0;

        Move(pDataBuffer^[0], pBuffer^[SizeOf(TChunkRec)], FFileSpec.RecLen);

        DataLen := SizeOf(TChunkRec) + FFileSpec.RecLen;

        BtrieveCheck( BtrvId( bChunkUpdate, FPositionRec,
                              pBuffer^, DataLen,
                              FKeyBuffer, FIndexNo, pClientId^),
                      FDiskFileName );
    finally
        FreeMem(pBuffer, BufferSize);
    end;

    WriteObjectData(Stream);
end;
{--------------------------------------------------------------------------}

procedure TBtrieveFile.Delete;
var
    DataLen : Word;

begin
    DataLen := 0;

    BtrieveCheck( BtrvId( bDelete, FPositionRec, pDataBuffer^, DataLen,
                  FKeyBuffer, FIndexNo, pClientId^),
                  FDiskFileName );

end;

{--------------------------------------------------------------------------}

function TBtrieveFile.GetIndexName : String;
var
    Index : TOdbIndex;
begin
    Index := FObjectFile.Indices[FIndexNo + 1] as TOdbIndex;

    Result := Index.Name;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.SetIndexName( const InIndexName : String );
var
    IndexNo  : Integer;
    Index    : TOdbIndex;
    Position : LongInt;
begin
    IndexNo := -1;

    if FObjectFile.Indices.GotoFirst then
    repeat
        Inc(IndexNo);

        Index := FObjectFile.Indices.CurrentObj as TOdbIndex;

        if AnsiCompareText(Index.Name, InIndexName) = 0 then
        begin
            FIndexNo := IndexNo;
            if RecordCount > 0 then
            begin
                Position := GetPosition;

                { MB Feb 15, 1998. Status 82 was occuring even with a
                  GetDirect to reestablish the key. Using a First operation
                  bypasses the problem. }
                First;
                GetDirect(Position);
            end;
            Exit;
        end;
    until not FObjectFile.Indices.GotoNext;

    { If reach here, InIndexName was not found. }
    RaiseOdbException(EOdbIndexDoesNotExist, InIndexName);
end;

{--------------------------------------------------------------------------}

function TBtrieveFile.LockCurrent : Boolean;
var
    Attempts  : Integer;
    Position  : LongInt;
    DataLen   : Word;
    Status    : SmallInt;

begin
    { assume failure }
    Result   := False;

    DataLen  := FFileSpec.RecLen;
    Position := GetPosition;

    if Position = -1 then
        Exit;

    Attempts := 0;
    while Attempts < CLockTries do
    begin
        Move(Position, pDataBuffer^, SizeOf(Position));

        Status := BtrvId( bGetDirect+bSingleNoWaitLock, FPositionRec,
                          pDataBuffer^, DataLen,
                          FKeyBuffer, FIndexNo, pClientId^);

        { bDataLengthError is normal when using chunks. }
        if Status = bDatalengthError then
            Status := bNoError;

        if (Status = bNoError) then
        begin
            Result := True;
            Exit;
        end;

        Inc(Attempts);

        OdbMisc.Delay(CBetweenLockDelayMS);
    end;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.UnLockCurrent;
var
    DataLen : Word;

begin
    DataLen   := 0;
    BtrieveCheck( BtrvId( bUnLock, FPositionRec, pDataBuffer^, DataLen,
                          FKeyBuffer, 0, pClientId^),
                  FDiskFileName );
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.UpdateKeyBuffer;
var
    CurSegment : Integer;
    CurIndex   : Integer;
    Position   : Integer;
    KeyLen     : Integer;

begin
    FillChar( FKeyBuffer, SizeOf(FKeyBuffer), 0 );

    CurSegment := 0;
    CurIndex   := 0;
    Position   := 1;

    repeat
        if CurIndex = FIndexNo then
        begin
            if (Position + FKeySpecs[CurSegment].KeyLen) <= SizeOf(FKeySpecs) then
            begin
                KeyLen := FKeySpecs[CurSegment].KeyLen;

                Move( pDataBuffer^[FKeySpecs[CurSegment].KeyPos-1],
                      FKeyBuffer[Position],
                      KeyLen );
            end;

            Inc( Position, FKeySpecs[CurSegment].KeyLen );
        end;

        if (not AreFlagsSet(FKeySpecs[CurSegment].KeyFlags, bSegmented)) then
            Inc(CurIndex);

        Inc(CurSegment);

    until (FKeySpecs[CurSegment].KeyPos = 0) or
          (CurSegment >= FKeySpecCount) or
          (CurIndex > FIndexNo);
end;

{--------------------------------------------------------------------------}

function TBtrieveFile.First : Boolean;
var
    DataLen : Word;
    Status  : SmallInt;

begin
    DataLen  := FFileSpec.RecLen;

    Status   := BtrvId( bGetFirst, FPositionRec, pDataBuffer^, DataLen,
                        FKeyBuffer, FIndexNo, pClientId^);

    { bDataLengthError is normal when using chunks. }
    if Status = bDatalengthError then
        Status := bNoError;

    BtrieveCheck(Status, FDiskFileName);

    Result   := (Status = bNoError);
end;

{--------------------------------------------------------------------------}

function TBtrieveFile.Last : Boolean;
var
    DataLen : Word;
    Status  : SmallInt;

begin
    DataLen  := FFileSpec.RecLen;

    Status   := BtrvId( bGetLast, FPositionRec, pDataBuffer^, DataLen,
                        FKeyBuffer, FIndexNo, pClientId^);

    { bDataLengthError is normal when using chunks. }
    if Status = bDatalengthError then
        Status := bNoError;

    BtrieveCheck(Status, FDiskFileName);

    Result   := (Status = bNoError);
end;

{--------------------------------------------------------------------------}

function TBtrieveFile.Next : Boolean;
var
    DataLen : Word;
    Status  : SmallInt;

begin
    DataLen  := FFileSpec.RecLen;

    Status   := BtrvId( bGetNext, FPositionRec, pDataBuffer^, DataLen,
                        FKeyBuffer, FIndexNo, pClientId^);

    { bDataLengthError is normal when using chunks. }
    if Status = bDatalengthError then
        Status := bNoError;

    BtrieveCheck(Status, FDiskFileName);

    Result   := (Status = bNoError);
end;

{--------------------------------------------------------------------------}

function TBtrieveFile.Prior : Boolean;
var
    DataLen : Word;
    Status  : SmallInt;

begin
    DataLen  := FFileSpec.RecLen;

    Status   := BtrvId( bGetPrev, FPositionRec, pDataBuffer^, DataLen,
                        FKeyBuffer, FIndexNo, pClientId^);

    { bDataLengthError is normal when using chunks. }
    if Status = bDatalengthError then
        Status := bNoError;

    BtrieveCheck(Status, FDiskFileName);

    Result   := (Status = bNoError);
end;

{--------------------------------------------------------------------------}

function TBtrieveFile.FindExact( const Fields : TUnOrderedList ) : Boolean;
var
    DataLen : Word;
    Status  : SmallInt;
begin
    TransferFieldsToBuffer(Fields, False);
    UpdateKeyBuffer;

    DataLen := FFileSpec.RecLen;

    Status  := BtrvId( bGetEqual, FPositionRec, pDataBuffer^, DataLen,
                       FKeyBuffer, FIndexNo, pClientId^);

    { bDataLengthError is normal when using chunks. }
    if Status = bDatalengthError then
        Status := bNoError;

    BtrieveCheck( Status, FDiskFileName );

    Result := (Status = bNoError);
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.FindNearest( const Fields : TUnOrderedList );
var
    DataLen : Word;
    Status  : SmallInt;
begin
    TransferFieldsToBuffer(Fields, False);
    UpdateKeyBuffer;

    DataLen := FFileSpec.RecLen;

    Status  := BtrvId( bGetGreatEqual, FPositionRec, pDataBuffer^, DataLen,
                       FKeyBuffer, FIndexNo, pClientId^);

    { bDataLengthError is normal when using chunks. }
    if Status = bDatalengthError then
        Status := bNoError;

    { MB Mar 06, 2001. If the key is greater than all records (ie, EOF is
      reached, move to the last record }
    if Status = bEOF then
       Self.Last;

    BtrieveCheck( Status, FDiskFileName );
end;

{--------------------------------------------------------------------------}

function TBtrieveFile.GetPosition : LongInt;
var
    DataLen  : Word;
    Status   : SmallInt;
begin
    DataLen := 4;

    Status := BtrvId( bGetPosition, FPositionRec, Result, DataLen,
                      FKeyBuffer, 0, pClientId^ );

    if Status = bPositionNotSet then
    begin
        Result := -1;
        Status := bNoError;
    end;

    BtrieveCheck( Status, FDiskFileName);
end;

{--------------------------------------------------------------------------}

function TBtrieveFile.GetDirect( const Position : LongInt ) : Boolean;
var
    DataLen : Word;
    Status  : SmallInt;
begin
    Result := False;
    if Position  = -1 then
        Exit;

    { make sure the data length is large enough to read the fixed portion
      of the record, in order for the ObjectSize property to return the
      property value. }
    DataLen   := FFileSpec.RecLen;
    Move(Position, pDataBuffer^, DataLen);

    Status    := BtrvId( bGetDirect, FPositionRec, pDataBuffer^, DataLen,
                         FKeyBuffer, FIndexNo, pClientId^);

    { bDataLengthError is normal when using chunks. }
    if Status = bDatalengthError then
        Status := bNoError;

    BtrieveCheck(Status, FDiskFileName);

    Result := (Status = bNoError);
end;

{--------------------------------------------------------------------------}

function TBtrieveFile.GetBookmark : Pointer;
begin
    New(TpBtrieveBookmark(Result));

    TpBtrieveBookmark(Result)^.IndexNo   := FIndexNo;
    TpBtrieveBookmark(Result)^.Position  := GetPosition;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.GotoBookmark( const Bookmark : Pointer );
begin
    FIndexNo   := TpBtrieveBookmark(Bookmark)^.IndexNo;
    GetDirect(TpBtrieveBookmark(Bookmark)^.Position);
end;

{--------------------------------------------------------------------------}

procedure TBtrieveFile.FreeBookmark( const Bookmark : Pointer );
begin
    Dispose(TpBtrieveBookmark(Bookmark));
end;

{--------------------------------------------------------------------------}

function TBtrieveFile.RecordCount : LongInt;
begin
    ReadAndCompareStats(False);
    Result := FFileSpec.NoRecords;
end;

{--------------------------------------------------------------------------}

function DefaultBtrieveStatusFunc( const Status : Integer ) : String; far;
begin
    Result := '';
    case Status of
        0001: Result := 'The operation paramater is invalid';
        0002: Result := 'The application encountered an I/O error';
        0003: Result := 'The file is not open';
        0004: Result := 'The application cannot find the key value';
        0005: Result := 'The record has a key field containing a duplicate key value';
        0006: Result := 'The key number parameter is invalid';
        0007: Result := 'The key number has changed';
        0008: Result := 'The current positioning is invalid';
        0009: Result := 'The operation encountered the end-of-file';
        0010: Result := 'The key field is not modifiable';
        0011: Result := 'The specified file name is invalid';
        0012: Result := 'The Microkernel cannot find the specified file';
        0013: Result := 'The Microkernel could not open the extension file for an extended file';
        0014: Result := 'The Microkernel cannot create or open the preimage file';
        0015: Result := 'The application encountered an I/O error during preimaging';
        0016: Result := 'The application encountered an expansion error';
        0017: Result := 'The application encountered a close error';
        0018: Result := 'The disk is full';
        0019: Result := 'The application encountered an unrecoverable error';
        0020: Result := 'The Microkernel or Btrieve Requester is inactive';
        0021: Result := 'The key buffer parameter is too short';
        0022: Result := 'The data buffer parameter is too short';
        0023: Result := 'The position block parameter is not 128 bytes long';
        0024: Result := 'The page size is invalid';
        0025: Result := 'The application cannot create the specified file';
        0026: Result := 'The number of keys specified is invalid';
        0027: Result := 'The key position is invalid';
        0028: Result := 'The record length is invalid';
        0029: Result := 'The key length is invalid';
        0030: Result := 'The file specified is not a Btrieve compatible file';
        0031: Result := 'The file is already extended';
        0032: Result := 'The file cannot be extended';
        0033: Result := 'The Microkernel cannot unload';
        0034: Result := 'The specified extension name is invalid';
        0035: Result := 'The application encountered a directory error';
        0036: Result := 'The application encountered a transaction error';
        0037: Result := 'Another transaction is active';
        0038: Result := 'The Microkernel encountered a transaction control file I/O error';
        0039: Result := 'A Begin Transaction operation must precede an End/Abort Transaction operation';
        0040: Result := 'The file access request exceeds the maximum number of files allowed';
        0041: Result := 'The Microkernel does not allow the attempted operation';
        0042: Result := 'A file previously opened in Accelerated mode was not closed';
        0043: Result := 'The specified record address is invalid';
        0044: Result := 'The specified key path is invalid';
        0045: Result := 'The specified key flags are  invalid';
        0046: Result := 'Access to the requested file is denied';
        0047: Result := 'The number of files opened exceeds the maximum allowed';
        0048: Result := 'The alternating collating sequence definition is invalid';
        0049: Result := 'The extended key type is invalid';
        0050: Result := 'The file owner is already set';
        0051: Result := 'The owner name is invalid';
        0052: Result := '';
        0053: Result := 'The language interface version invalid';
        0054: Result := 'The variable-length portion of the record is corrupt';
        0055: Result := 'The application specifed an invalid attribute for an autoincrement key';
        0056: Result := 'An index is incomplete';
        0057: Result := '';
        0058: Result := 'The compression buffer length is too short';
        0059: Result := 'The specified file already exists';
        0060: Result := 'The specified reject count has been reached';
        0061: Result := 'The work space is too small';
        0062: Result := 'The description is incorrect';
        0063: Result := 'The data buffer parameter specified on an Insert Extended operation is invalid';
        0064: Result := 'The filter limit has been reached';
        0065: Result := 'The field offset is incorrect';
        0066: Result := 'The maximum number of opened databases has been exceeded';
        0067: Result := 'The Microkernel cannot open the SQL data dictionary files';
        0068: Result := 'The Microkernel cannot perform the RI Delete Cascase operation';
        0069: Result := 'The Delete operation specified a file that is damaged';
        0070: Result := '';
        0071: Result := 'The is a violation of the RI definitions';
        0072: Result := 'The Microkernel cannot open the RI referenced file';
        0073: Result := 'The RI definition is out of sync';
        0074: Result := 'The Microkernel aborted the transaction';
        0075: Result := '';
        0076: Result := 'There is a conflict on the referenced file';
        0077: Result := 'The application encountered a wait error';
        0078: Result := 'The Microkernel detected a deadlock condition';
        0079: Result := '';
        0080: Result := 'The Microkernel encountered a record-level conflict';
        0081: Result := 'The Microkernel encountered a lock error';
        0082: Result := 'The Microkernel lost positioning';
        0083: Result := 'The Microkernel attempted to read a record outside the transaction';
        0084: Result := 'The record or page is locked';
        0085: Result := 'The file is locked';
        0086: Result := 'The file table is full';
        0087: Result := 'The handle table is full';
        0088: Result := 'The application encountered an incompatible mode error';
        0089: Result := '';
        0090: Result := 'Redirected device table full';
        0091: Result := 'The application encountered a server error';
        0092: Result := 'The transaction table is full';
        0093: Result := 'The record lock types are incompatible';
        0094: Result := 'The aplication encountered a permission error';
        0095: Result := 'The session is no longer valid';
        0096: Result := 'A communications environment error occured';
        0097: Result := 'The data buffer is too small';
        0098: Result := 'The Microkernel detected an internal transaction error';
        0099: Result := 'Btrieve requester unable to access NetWare Runtime server';
        0100: Result := 'No cache buffers are available';
        0101: Result := 'Insufficient operating system memory is available';
        0102: Result := 'Insufficient stack space is available';
        0103: Result := 'The chunk offset is too big';
        0104: Result := 'The Microkernel does not recognize the locale';
        0105: Result := 'The file cannot be created with Variable-tail Allocation Tables (VATs)';
        0106: Result := 'The Microkernel cannot perform a Get Next Chunk operation';
        0107: Result := 'The application attempted to perform a chunk operation on a pre-v6.0 file';
        0108: Result := '';
        0109: Result := 'An unknown error was encountered either creating or accessing a semaphore';
        0130: Result := 'The Microkernel ran out of system locks';
        0132: Result := 'File full';
        0133: Result := 'More than 5 concurrent users attempted to access the same Btrieve compatible file';
        1001: Result := 'The lock parameter is out of range';
        1002: Result := 'A memory allocation error occurred';
        1003: Result := 'An invalid memory size parameter was specified';
        1004: Result := 'A page size error occurred';
        1005: Result := 'The preimage drive parameter is invalid';
        1006: Result := 'The preimage buffer parameter is out of range';
        1007: Result := 'The files parameter is out of range';
        1008: Result := 'The initialization parameter is invalid';
        1009: Result := 'The transaction filename parameter is invalid';
        1010: Result := 'An error occurred during the access of the transaction control file';
        1011: Result := 'The compression buffer parameter is out of range';
        1012: Result := 'Invalid /n: option (pre v6.x Windows workstation engine)';
        1013: Result := 'The task list is full';
        1014: Result := 'Stop warning. The application still has files or transactions active';
        1015: Result := 'One of the pointer parameters passed to the Microkernel is invalid';
        1016: Result := 'The Microkernel is already initialized';
        1017: Result := 'The Btrieve requester is unable to find the resource file WBTRVRES.DLL';
        1018: Result := 'The application attempted to call the Microkernel from a Btrieve callback function';
        1019: Result := 'The Microkernel cancelled the current Btrieve operation at the request of the ' +
                        'application''s callback function';
        1020: Result := 'Btrieve requester interface communication error';
        2001: Result := 'The memory allocation is insufficient';
        2002: Result := 'The option is invalid or out of range';
        2003: Result := 'The requester does not allow local access to the specified file';
        2004: Result := 'SPX is not installed';
        2005: Result := 'An incorrect version of SPX is installed';
        2006: Result := 'There is no available SPX connection';
        2007: Result := 'A ponter parameter is invalid';
    end; { case }
end;

begin
    BtrieveStatusFunc := DefaultBtrieveStatusFunc;
end.
